# Errors
# - BlockNotFoundError
# - UnknownOperationError


class BlockNotFoundError(Exception):
    pass

class UnknownOperationError(Exception):
    pass

class FileNotFoundError(Exception):
    pass
